<?php

namespace Drupal\commerce_nzpost\Packer;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_order\Entity\OrderItemInterface;
use Drupal\commerce_shipping\Packer\DefaultPacker;
use Drupal\commerce_shipping\Packer\PackerInterface;
use Drupal\commerce_shipping\ProposedShipment;
use Drupal\commerce_shipping\ShipmentItem;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\physical\Weight;
use Drupal\physical\WeightUnit;
use Drupal\physical\VolumeUnit;
use Drupal\physical\LengthUnit;
use Drupal\profile\Entity\ProfileInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

/**
 * Defines a shipment packer for NzPost.
 *
 * @package Drupal\commerce_nzpost\Packer
 */
class CommerceNzPostPacker extends DefaultPacker implements PackerInterface {

  use StringTranslationTrait;

  const VOLUME_UNIT_OF_MEASURE = VolumeUnit::CUBIC_MILLIMETER;
  const LENGTH_UNIT_OF_MEASURE = LengthUnit::MILLIMETER;
  const WEIGHT_UNIT_OF_MEASURE = WeightUnit::KILOGRAM;

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * Config for this module.
   *
   * @var \Drupal\Core\Config\ConfigFactory
   */
  protected $config;

  /**
   * Maximum size, in mm of a single package.
   *
   * @var int
   */
  protected $maxSize;

  /**
   * The longest a single side of a package can be in mm.
   *
   * @var int
   */
  protected $maxSingleSide;

  /**
   * The maximum weight a single package can be in kg.
   *
   * @var float
   */
  protected $maxWeight;

  /**
   * Constructs a new DefaultPacker object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher
   *   The event dispatcher.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, EventDispatcherInterface $event_dispatcher) {
    $this->entityTypeManager = $entity_type_manager;
    $this->eventDispatcher = $event_dispatcher;

    return $this;
  }

  /**
   * Get module config.
   *
   * @param array $configuration
   *   Config.
   */
  public function setConfig(array $configuration) {
    $this->maxWeight = $configuration['max_weight'];
    $this->maxSize = $configuration['max_size'];
    $this->maxSingleSide = $configuration['max_single_side'];
  }

  /**
   * {@inheritdoc}
   */
  public function pack(OrderInterface $order, ProfileInterface $shipping_profile) {

    // A counter: one shipment == one box == one shipmentId that needs shipping.
    $shipmentId = 0;
    $shipments = [];
    $shipments[$shipmentId]['volume'] = 0;
    $shipments[$shipmentId]['total_weight'] = 0;
    $shipments[$shipmentId]['value'] = 0;
    foreach ($order->getItems() as $order_item) {
      $purchased_entity = $order_item->getPurchasedEntity();

      // Ship only shippable purchasable entity types.
      if (!$purchased_entity || !$purchased_entity->hasField('weight')) {
        continue;
      }

      $quantity = $order_item->getQuantity();

      // @TODO this logic attempts to fit all items of a single product
      // into one package eg 500 tubes on 85ml will NOT be split and will not be
      // shippable. NEEDS FIXING.
      if (!$this->addItemToShipment($order_item, $shipments[$shipmentId])) {
        $shipmentId++;
        $shipments[$shipmentId]['volume'] = 0;
        $shipments[$shipmentId]['total_weight'] = 0;
        $shipments[$shipmentId]['value'] = 0;
        if (!$this->addItemToShipment($order_item, $shipments[$shipmentId])) {
          // Adding has completely failed.
          return [];
          // @TODO throw error? With useful details to help track it down.
        }
      }
      $shipments[$shipmentId]['volume'] += $this->getVolume($order_item);
      $shipments[$shipmentId]['total_weight'] += $this->getWeight($order_item)->getNumber();
      $shipments[$shipmentId]['value'] += $order_item->getAdjustedTotalPrice()->getNumber();
    }

    $proposed_shipments = [];

    foreach ($shipments as $key => $shipment) {
      if (!empty($shipment['items'])) {
        $proposed_shipments[] = new ProposedShipment([
          'type' => $this->getShipmentType($order),
          'order_id' => $order->id(),
          'title' => 'Shipment ' . $key,
          'items' => $shipment['items'],
          'shipping_profile' => $shipping_profile,
          'custom_fields' => [
            'dimensions' => $this->getPackageDimensions($shipment['volume']),
            'value' => $shipment['value'],
            'weight' => $shipment['total_weight'],
          ],
        ]);
      }
    }

    return $proposed_shipments;
  }

  /**
   * Gets the weight of the order item.
   *
   * The weight will be empty if the shippable trait was added but the existing
   * entities were not updated. Add 1 gram as nzpost doesn't support 0 weights.
   *
   * @param \Drupal\commerce_order\Entity\OrderItemInterface $order_item
   *   The order item.
   *
   * @return \Drupal\physical\Weight
   *   The weight.
   */
  protected function getWeight(OrderItemInterface $order_item) {
    $purchasedEntity = $order_item->getPurchasedEntity();

    if ($purchasedEntity->get('weight')->isEmpty()) {
      $weight = new Weight(1, WeightUnit::GRAM);
    }
    else {
      /** @var \Drupal\physical\Plugin\Field\FieldType\MeasurementItem $weight_item */
      $weight_item = $purchasedEntity->get('weight')->first();

      $weight = $weight_item->toMeasurement();
    }

    return $weight;
  }

  /**
   * Add item to provided shipment.
   *
   * @param \Drupal\commerce_order\Entity\OrderItemInterface $order_item
   *   The order item to add.
   * @param array $shipment
   *   The shipment to add the order to.
   *
   * @return bool
   *   Successful add or not?
   */
  protected function addItemToShipment(OrderItemInterface $order_item, array &$shipment) {
    // We care about:
    // - weight first ( cheaper calculation )
    // - volume second ( slightly more expensive calculation )
    // Will adding this item put us over the max weight we are allowed?
    $item_weight = $this->getWeight($order_item)->multiply($order_item->getQuantity())->getNumber();
    if (($item_weight + $shipment['total_weight']) < $this->maxWeight) {
      // Now check the total girth of this shipment using volume.
      $item_volume = $this->getVolume($order_item);
      $proposed_volume = $shipment['volume'] + $item_volume;

      // To calculate our dimensions we:
      // - take the total volume of the shipment + new item
      // - divide by the maxSingleSide
      // - root the result ( giving us lengths for the other two sides
      // - multiply that result by 2, add on the maxSingleSide and we have our
      // package 'size' and can compare that to the max NZ Post will allow.
      if ((sqrt($proposed_volume / $this->maxSingleSide) * 2 + $this->maxSingleSide) < $this->maxSize) {
        // Go ahead and add this item, update shipment values at the same time.
        $shipment['items'][] = new ShipmentItem([
          'order_item_id' => $order_item->id(),
          'title' => $order_item->getPurchasedEntity()->getOrderItemTitle(),
          'quantity' => $order_item->getQuantity(),
          'weight' => $this->getWeight($order_item)->multiply($order_item->getQuantity()),
          'declared_value' => $this->getDeclaredValue($order_item),
        ]);
        return TRUE;
      }
    }
    return FALSE;
  }

  /**
   * Returns the total volume for this order_item in mm.
   *
   * Eg given a quantity of 3 and:.
   *  - length: 10mm
   *  - width: 10mm
   *  - height: 10mm
   *  - quantity: 3
   *
   * Return = 3000mm.
   *
   * @param \Drupal\commerce_order\Entity\OrderItemInterface $order_item
   *   The item we need to pack.
   *
   * @return float
   *   return value in cubic mm.
   */
  private function getVolume(OrderItemInterface $order_item) {
    $total_volume = NULL;
    $purchased_entity = $order_item->getPurchasedEntity();
    if ($purchased_entity->hasField('dimensions') && !$purchased_entity->get('dimensions')->isEmpty()) {
      /** @var \Drupal\physical\Plugin\Field\FieldType\DimensionsItem $dimensions */
      $dimensions = $purchased_entity->get('dimensions')->first();

      $volume = $dimensions
        ->getHeight()
        ->convert(SELF::LENGTH_UNIT_OF_MEASURE)
        ->multiply($dimensions->getWidth()->convert(SELF::LENGTH_UNIT_OF_MEASURE)->getNumber())
        ->multiply($dimensions->getLength()->convert(SELF::LENGTH_UNIT_OF_MEASURE)->getNumber())
        ->getNumber();

      $total_volume += (float) $volume * $order_item->getQuantity();
    }
    return $total_volume;
  }

  /**
   * Get the total value for this item.
   */
  private function getDeclaredValue(OrderItemInterface $order_item) {
    return $order_item->getTotalPrice();
  }

  /**
   * This works out the lengths of the 3 sides of a package.
   *
   * The specific lengths of the sides do not matter as they do not affect
   * pricing, so we can pretend that the longest side we have is the longest
   * available and that the two other sides are of equal length.  Later we can
   * send this in our lookup to the NZ Post API.
   *
   * @param float $volume
   *   Volume of Package in mm.
   *
   * @return array
   *   Keyed by length, width, thickness.
   */
  private function getPackageDimensions(float $volume) {
    $short_side = sqrt($volume / $this->maxSingleSide);
    $ret['length'] = $this->maxSingleSide;
    $ret['height'] = $short_side;
    $ret['thickness'] = $short_side;
    return $ret;
  }

}
