<?php

namespace Drupal\commerce_nzpost\Plugin\Commerce\ShippingMethod;

use Drupal\commerce_price\Price;
use Drupal\commerce_shipping\Entity\ShipmentInterface;
use Drupal\commerce_shipping\PackageTypeManagerInterface;
use Drupal\commerce_shipping\Plugin\Commerce\ShippingMethod\ShippingMethodBase;
use Drupal\commerce_shipping\ShippingRate;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Form\FormStateInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Drupal\commerce_nzpost\Packer\CommerceNzPostPacker as Packer;
use Drupal\commerce_nzpost\RateLookupService;

/**
 * Provides the NzPost shipping method.
 *
 * @CommerceShippingMethod(
 *   id = "nzpost",
 *   label = @Translation("NZ Post"),
 *   services = {
 *     "TIALP" = @Translation("International Air Large Package"),
 *     "TIASPC" = @Translation("International Air Small Package"),
 *     "TIEC" = @Translation("International Economy Courier"),
 *     "TIEX" = @Translation("International Express Courier"),
 *   }
 * )
 */
class NzPost extends ShippingMethodBase {

  /**
   * Constant for Domestic Shipping. NOT IMPLEMENTED AT PRESENT.
   */
  const SHIPPING_TYPE_DOMESTIC = 'domestic';

  /**
   * Constant for International Shipping.NOT IN USE, EVERYTHING'S INTERNATIONAL.
   */
  const SHIPPING_TYPE_INTERNATIONAL = 'intl';

  /**
   * The event dispatcher.
   *
   * @var \Symfony\Component\EventDispatcher\EventDispatcherInterface
   */
  protected $eventDispatcher;

  /**
   * The Service Plugins.
   *
   * @var \Drupal\Core\Plugin\DefaultLazyPluginCollection
   */
  protected $plugins;

  /**
   * Commerce NzPost Logger Channel.
   *
   * @var \Psr\Log\LoggerInterface
   */
  protected $watchdog;

  /**
   * The NZPost Connection.
   *
   * @var \Drupal\commerce_nzpost\NZPostServiceManager
   */
  protected $nzPostServiceManager;

  /**
   * Shipping services.
   *
   * @var \Drupal\commerce_shipping\ShippingService
   */
  protected $services;


  /**
   * NZPost rate lookup service.
   *
   * @var \Drupal\commerce_nzpost\RateLookupService
   */
  protected $rateLookupService;

  /**
   * NZ Post packer.
   *
   * @var \Drupal\commerce_nzpost\Packer
   */
  protected $packer;

  /**
   * Constructs a new NzPost object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\commerce_shipping\PackageTypeManagerInterface $package_type_manager
   *   The package type manager.
   * @param \Symfony\Component\EventDispatcher\EventDispatcherInterface $event_dispatcher,
   *   Event Dispatcher, not used.
   * @param \Drupal\commerce_nzpost\RateLookupService $rate_lookup_service
   *   Rate Lookup Service.
   * @param \Drupal\commerce_nzpost\Packer\CommerceNzPostPacker $packer
   *   NZ Post Packer.
   */
  public function __construct(array $configuration,
                              $plugin_id,
                              $plugin_definition,
                              PackageTypeManagerInterface $package_type_manager,
                              EventDispatcherInterface $event_dispatcher,
                              RateLookupService $rate_lookup_service,
                              Packer $packer) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $package_type_manager);

    $this->rateLookupService = $rate_lookup_service;
    $this->packer = $packer;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('plugin.manager.commerce_package_type'),
      $container->get('event_dispatcher'),
      $container->get('commerce_nzpost.ratelookup'),
      $container->get('commerce_nzpost.commerce_nzpost_packer')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {

    return [
      'api_information' => [
        'api_key' => '',
      ],
      'package_information' => [
        'max_weight' => 20,
        'max_single_side' => 1050,
        'max_size' => 2000,
      ],
      'plugins' => [],
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    // Select all services by default.
    if (empty($this->configuration['services'])) {
      $service_ids = array_keys($this->services);
      $this->configuration['services'] = array_combine($service_ids, $service_ids);
    }

    $form['package_information'] = [
      '#type' => "details",
      '#title' => $this->t("Package sizing information"),
      '#description' => $this->t("Customise your NZ Post Package size information, you need to ensure you don't exceed NZ Post limits or you will get incorrect or missing estimates from the API."),
      '#weight' => $this->isConfigured() ? 5 : -5,
      '#open' => !$this->isConfigured(),
    ];

    $form['package_information']['max_weight'] = [
      '#type' => 'number',
      '#title' => $this->t('Max Package Weight in Kg'),
      '#description' => $this->t('Enter the maximum package weight'),
      '#default_value' => $this->configuration['package_information']['max_weight'],
      '#required' => TRUE,
    ];

    $form['package_information']['max_single_side'] = [
      '#type' => 'number',
      '#title' => $this->t('Max Longest Side in mm'),
      '#description' => $this->t('Enter the longest a single side available for packages.'),
      '#default_value' => $this->configuration['package_information']['max_single_side'],
      '#required' => TRUE,
    ];

    $form['package_information']['max_size'] = [
      '#type' => 'number',
      '#title' => $this->t('Max size in mm ( L + W + H )'),
      '#description' => $this->t('Enter the maximum summed dimensions for a package.'),
      '#default_value' => $this->configuration['package_information']['max_size'],
      '#required' => TRUE,
    ];

    $form['api_information'] = [
      '#type' => 'details',
      '#title' => $this->t('API information'),
      '#description' => $this->isConfigured() ? $this->t('Update your NZ Post API information.') : $this->t('Fill in your NZPost API information.'),
      '#weight' => $this->isConfigured() ? 10 : -10,
      '#open' => !$this->isConfigured(),
    ];

    $form['api_information']['api_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('API key'),
      '#description' => $this->t('Enter your NZ Post API key.'),
      '#default_value' => $this->configuration['api_information']['api_key'],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {

    parent::validateConfigurationForm($form, $form_state);

    foreach ($this->nzPostServiceManager->getDefinitions() as $plugin_id => $definition) {
      if (!empty($form[$plugin_id])) {
        /** @var \Drupal\commerce_nzpost\Plugin\Commerce\ShippingMethod\NZPost $plugin */
        $plugin = $this->plugins->get($plugin_id);
        $plugin->validateConfigurationForm($form[$plugin_id], SubformState::createForSubform($form[$plugin_id], $form_state->getCompleteForm(), $form_state));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    if (!$form_state->getErrors()) {

      $values = $form_state->getValue($form['#parents']);

      $this->configuration['api_information']['api_key'] = $values['api_information']['api_key'];
      $this->configuration['package_information']['max_weight'] = $values['package_information']['max_weight'];
      $this->configuration['package_information']['max_single_side'] = $values['package_information']['max_single_side'];
      $this->configuration['package_information']['max_size'] = $values['package_information']['max_size'];
    }
    parent::submitConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function calculateRates(ShipmentInterface $shipment) {

    $availableRates = [];

    if ($shipment->getShippingProfile()->address->isEmpty()) {
      return [];
    }

    // To ship an order we need to:
    // - Package it into then minimum number of boxes that are shippable.
    // - Request rates from NZ Post API for each package.
    // - Return shipping quotes that are available for every package.
    $this->rateLookupService->setConfig($this->configuration);
    $rates = $this->rateLookupService->getRates($shipment, $this->configuration);

    foreach ($rates as $key => $rate) {
      $price = new Price($rate['price_including_gst'], 'NZD');
      if (array_key_exists($key, $this->getServices())) {
        $availableRates[] = new ShippingRate($key, $this->services[$key], $price);
      }
    }

    return $availableRates;
  }

  /**
   * Determine if we have the minimum information to connect to NZPost.
   *
   * @return bool
   *   TRUE if there is enough information to connect, FALSE otherwise.
   */
  protected function isConfigured() {
    $api_information = $this->configuration['api_information'];
    $package_information = $this->configuration['package_information'];

    if (!empty($api_information['api_key']) &&
      !empty($package_information['max_weight']) &&
      !empty($package_information['max_single_side']) &&
      !empty($package_information['max_size'])
    ) {
      return TRUE;
    }
    else {
      return FALSE;
    }

  }

}
