<?php

namespace Drupal\commerce_nzpost;

use GuzzleHttp\Client;
use Drupal\commerce_nzpost\Packer\CommerceNzPostPacker;
use Drupal\commerce_shipping\Entity\ShipmentInterface;

/**
 * Class RateLookupService.
 */
class RateLookupService implements RateLookupServiceInterface {

  const API_URL = 'https://api.nzpost.co.nz/ratefinder/international.json';
  /**
   * GuzzleHttp\Client definition.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * Entity\Order definition.
   *
   * @var \Drupal\commerce_order\Entity\Order
   */
  protected $order;

  /**
   * Drupal\commerce_nzpost\Packer\CommerceNzPostPacker definition.
   *
   * @var \Drupal\commerce_nzpost\Packer\CommerceNzPostPacker
   */
  protected $packer;

  /**
   * Config.
   *
   * @var array
   */
  protected $configuration;

  /**
   * Shipment.
   *
   * @var \Drupal\commerce_shipping\Entity\ShipmentInterface
   */
  protected $commerceShipment;

  /**
   * Constructs a new RateLookupService object.
   */
  public function __construct(Client $http_client, CommerceNzPostPacker $packer) {
    $this->httpClient = $http_client;
    $this->packer = $packer;
  }

  /**
   * Sets configuration for requests.
   *
   * @param array $configuration
   *   A configuration array from a CommerceShippingMethod.
   */
  public function setConfig(array $configuration) {
    $this->configuration = $configuration;
  }

  /**
   * Set the shipment for rate requests.
   *
   * @param \Drupal\commerce_shipping\Entity\ShipmentInterface $commerce_shipment
   *   A Drupal Commerce shipment entity.
   */
  public function setShipment(ShipmentInterface $commerce_shipment) {
    $this->commerceShipment = $commerce_shipment;
  }

  /**
   * Get rates available for this order.
   *
   * @param \Drupal\commerce_shipping\Entity\ShipmentInterface $shipment
   *   The shipment.
   * @param array $config
   *   General config.
   *
   * @return array
   *   All rates that are available for all the packages. Prices are summed.
   */
  public function getRates(ShipmentInterface $shipment, array $config) {
    // Pack the order and get packages.
    // Make request to NZ Post API for each package.
    // All rates returned by NZ POST API.
    $availableRates = [];
    // All rates for all packages keyed numerically in order.
    $packages_rates = [];

    if ($shipment->getShippingProfile()->get('address')->isEmpty()) {
      return [];
    }

    $country_code = $shipment->getShippingProfile()->get('address')->first()->getCountryCode();
    /*
     * There is currently no way to do a 'NOT' selection in Commerce Shipping.
     * Ideally we would only invoke this shipping method when the shipping
     * country is 'NOT NZ', but we can't do that in the UI, so we do it here
     * instead.
     */
    if ($country_code == 'NZ') {
      return [];
    }

    $this->packer->setConfig($this->configuration['package_information']);
    $packages = $this->packer->pack($shipment->getOrder(), $shipment->getShippingProfile());
    $api_key = $this->configuration['api_information']['api_key'];
    foreach ($packages as $package) {
      $request = $this->httpClient->get(SELF::API_URL, [
        'query' => [
          'api_key' => $api_key,
          'country_code' => $country_code,
          'value' => $package->getCustomFields()['value'],
          'length' => $package->getCustomFields()['dimensions']['length'],
          'height' => $package->getCustomFields()['dimensions']['height'],
          'thickness' => $package->getCustomFields()['dimensions']['thickness'],
          'weight' => $package->getCustomFields()['weight'],
        ],
      ]);

      $response = json_decode($request->getBody(), TRUE);

      if (isset($response['products'])) {
        $packages_rates[] = $this->parseRates($response);
      }
    }

    foreach ($packages_rates as $key => $package_rates) {
      // On first loop we need to add ALL the rates available for this package,
      // this is our baseline and we eliminate from there, we never add
      // another service after the first package, we only remove and add to the
      // cost.
      if ($key === 0) {
        foreach ($package_rates as $k => $rate) {
          $availableRates[$k]['price'] = $rate['price'];
          $availableRates[$k]['price_including_gst'] = $rate['price_including_gst'];
        }
      }
      else {
        foreach ($package_rates as $k => $rate) {
          // Track the services for this package so we can remove those
          // that are currently listed but need to be taken out.
          // We can just ignore this service if it's not in the array already
          // as it wasn't available for the first package.
          foreach ($rate as $r) {
            if (array_key_exists($k, $availableRates)) {
              $availableRates[$k]['price'] += $rate['price'];
              $availableRates[$k]['price_including_gst'] += $rate['price_including_gst'];
              $added_services[$k] = $k;
            }
          }
          // Remove services that are currently in the master array but weren't
          // available on the current package.
          $availableRates = array_intersect_key($availableRates, $added_services);
        }
      }
    }

    return $availableRates;
  }

  /**
   * Turns API response into more readable array.
   *
   * @param array $response
   *   Array of json decoded data from the NZ Post API.
   *
   * @return array
   *   Handy formatted array of available rates for a package.
   */
  private function parseRates(array $response) {
    $ret = [];

    if (count($response['products'])) {
      foreach ($response['products'] as $r) {
        $service_code = $r['service_code'];
        $ret[$service_code] = $r;
        $ret[$service_code]['price'] = $r['price'];
        $ret[$service_code]['price_including_gst'] = $r['price_including_gst'];
      }
    }
    return $ret;
  }

}
